import time
from io import BytesIO
from typing import Dict, List, Tuple
import geopandas as gpd

import numpy as np
import requests
from PIL import Image, ImageDraw
from scipy.ndimage import binary_fill_holes

from ._utils import get_YX_area, XY2deg

__all__ = ["download_sateimgs", "download_all_tiles"]


def _download_one_tile(args):
    """
    Download the tile for the given x and y coordinates.
    From Esri living atlas, the token is required to download the tiles.

    Depending on the API of the tile server.
    """
    # args
    base_url, Z, X, Y = args
    base_url = base_url.rstrip("/")

    # the url of the tile
    # base_url = "https://wayback.maptiles.arcgis.com/arcgis/rest/services/World_Imagery/WMTS/1.0.0/default028mm/MapServer/tile/41468/15/"
    # url = base_url + str(Y) + "/" + str(X) + "?token=123"
    url = f"{base_url}/{Z}/{Y}/{X}"

    # the filename of the tile
    filename = str(Y) + "_" + str(X)

    try:
        try:
            response = requests.get(url)
            response.raise_for_status()
            img_bytes = BytesIO(response.content)
            return filename, img_bytes
        except:
            time.sleep(1)
            response = requests.get(url)
            response.raise_for_status()
            img_bytes = BytesIO(response.content)
            return filename, img_bytes
    except:
        return filename


def download_all_tiles(
    base_url: str, Z: int, Y_X: List[str]
) -> Tuple[Dict[str, Image.Image], List[str]]:
    """
    Download all the tiles for the given x and y coordinates.
    From Esri living atlas, the token is required to download the tiles.

    The tiles are downloaded into the given directory.

    Args:
    - base_url (str): the base url of the tile server.
    - Z (int): the zoom level.
    - Y_X (List[str]): the list of y_x coordinates.

    Returns:
    - cached_tiles (Dict[str, Image.Image]): the cached tiles.
    - fail_tile_list (List[str]): the list of failed tiles.
    """

    # the arguments for the download_one_tile function
    tile_args = [
        (base_url, int(Z), int(X), int(Y)) for Y, X in [y_x.split("_") for y_x in Y_X]
    ]

    # download the tiles
    cached_tiles = {}
    fail_tile_list = []

    # sequential version
    for arg in tile_args:
        result = _download_one_tile(arg)
        if isinstance(result, tuple):
            tile_no, img = result
            cached_tiles[tile_no] = Image.open(img)
        elif isinstance(result, str):
            fail_tile_list.append(result)

    return cached_tiles, fail_tile_list


def _concat_img_one_region(args):
    """
    Concatenate the tiles to obtain the regional satellite images.
    """
    # args
    region, tiles, files = args

    # the region should not be too large
    if len(tiles) < 20000:
        # get the x and y coordinates of the tiles
        ys, xs = zip(*[(int(i.split("_")[0]), int(i.split("_")[1])) for i in tiles])

        # get the min and max x and y coordinates
        row_min, col_min, row_max, col_max = min(ys), min(xs), max(ys), max(xs)

        # generate the mask
        mask_width = col_max - col_min + 1
        mask_height = row_max - row_min + 1

        mask_whole = np.zeros([mask_height * 256, mask_width * 256, 3], dtype=np.int32)

        # the pixel coordinates
        pixel_coords = []
        for tile in tiles:
            try:
                # read the tile
                img_temp = files[tile]

                # get the tile's x and y coordinates
                row, col = [int(x) for x in tile.split("_")]

                # convert the tile's x and y coordinates to the longitude and latitude
                minlat, minlon = XY2deg(col, row, 15)
                maxlat, maxlon = XY2deg(col + 1, row + 1, 15)

                if region.iloc[0].geometry.geom_type != "MultiPolygon":
                    # if the region is not a multi-polygon
                    for coord in region.iloc[0].geometry.exterior.coords:
                        # convert the longitude and latitude to the pixel coordinates, offset
                        x = (coord[0] - minlon) * (256 / (maxlon - minlon))
                        y = (coord[1] - minlat) * (256 / (maxlat - minlat))

                        # convert the pixel coordinates to the regional pixel coordinates, overall
                        temp_x, temp_y = (
                            x + (col - col_min) * 256,
                            y + (row - row_min) * 256,
                        )
                        if temp_x > 256 * mask_width:
                            temp_x = 256 * mask_width
                        if temp_y > 256 * mask_height:
                            temp_y = 256 * mask_height
                        if temp_x < 0:
                            temp_x = 0
                        if temp_y < 0:
                            temp_y = 0

                        # store the pixel coordinates
                        pixel_coords.append((temp_x, temp_y))
                else:
                    # if the region is a multi-polygon
                    for polygon in region.iloc[0].geometry.geoms:
                        for coord in polygon.exterior.coords:
                            # convert the longitude and latitude to the pixel coordinates, offset
                            x = (coord[0] - minlon) * (256 / (maxlon - minlon))
                            y = (coord[1] - minlat) * (256 / (maxlat - minlat))

                            # convert the pixel coordinates to the regional pixel coordinates, overall
                            temp_x, temp_y = (
                                x + (col - col_min) * 256,
                                y + (row - row_min) * 256,
                            )
                            if temp_x > 256 * mask_width:
                                temp_x = 256 * mask_width
                            if temp_y > 256 * mask_height:
                                temp_y = 256 * mask_height

                            if temp_x < 0:
                                temp_x = 0
                            if temp_y < 0:
                                temp_y = 0

                            # store the pixel coordinates
                            pixel_coords.append((temp_x, temp_y))

                mask_whole[
                    (row - row_min) * 256 : (row - row_min + 1) * 256,
                    (col - col_min) * 256 : (col - col_min + 1) * 256,
                    :,
                ] = img_temp
            except:
                continue

        # fill the holes in the mask
        if mask_whole.sum() != 0:
            # convert the mask to the image
            im = np.uint8(mask_whole)
            im_zero = Image.new("L", (mask_width * 256, mask_height * 256), 0)
            draw = ImageDraw.Draw(im_zero)
            for i in range(len(pixel_coords)):
                draw.line(
                    [pixel_coords[i], pixel_coords[(i + 1) % len(pixel_coords)]],
                    fill="red",
                    width=5,
                )
            filled_mask = binary_fill_holes(im_zero)

            # save the image
            im = im * filled_mask.reshape(filled_mask.shape[0], filled_mask.shape[1], 1)
            im = Image.fromarray(np.uint8(im))
            return int(region.index[0]), im
        else:
            return int(region.index[0])

    else:
        return int(region.index[0])


def _concatenate_tiles(
    region_list, area_shp: gpd.GeoDataFrame, cached_tiles: Dict[str, Image.Image]
):
    """
    Concatenate the tiles to obtain all the regional satellite images given the area.
    """

    # list of region idx
    region_list = sorted(region_list)

    # list of the arguments
    args = [
        (
            area_shp[area_shp.index == i],
            area_shp.loc[area_shp.index == i, "Y_X"].item(),
            {
                key: cached_tiles[key]
                for key in area_shp.loc[area_shp.index == i, "Y_X"].item()
                if key in cached_tiles
            },
        )
        for i in region_list
    ]

    # concatenate the tiles
    regional_imgs = {}
    fail_region_list = []

    # sequential version
    for arg in args:
        result = _concat_img_one_region(arg)
        if isinstance(result, tuple):
            idx, img = result
            regional_imgs[idx] = img
        else:
            fail_region_list.append(result)

    return regional_imgs, fail_region_list


def download_sateimgs(
    area_shp: gpd.GeoDataFrame,
    base_url: str = "https://wayback.maptiles.arcgis.com/arcgis/rest/services/World_Imagery/WMTS/1.0.0/default028mm/MapServer/tile/41468",
    Z: int = 15,
) -> Dict[int, Image.Image]:
    """
    Download the satellite images for the given area from Esri living atlas.

    Args:
    - area_shp (gpd.GeoDataFrame): the areas to download the satellite images.

    Returns:
    - regional_imgs (Dict[int, Image.Image]): the regional satellite images.
    """

    # coordinate system
    area_shp = area_shp.to_crs(epsg=4326)

    # get the x and y coordinates of the tiles
    area_shp, Y_X = get_YX_area(area_shp)

    # download the tiles
    cached_tiles: Dict[str, Image.Image] = {}
    remaining = Y_X
    while remaining != []:
        cached_tiles_tmp, fail_tile_list = download_all_tiles(base_url, Z, remaining)
        cached_tiles.update(cached_tiles_tmp)
        remaining = fail_tile_list

    # concatenate the tiles to obtain the regional satellite images
    regional_imgs = {}
    remaining = area_shp.index.to_list()
    while remaining != []:
        regional_imgs_tmp, fail_region_list = _concatenate_tiles(
            remaining,
            area_shp,
            cached_tiles,
        )
        regional_imgs.update(regional_imgs_tmp)
        remaining = fail_region_list

    return regional_imgs



from typing import Union

import cv2
import numpy as np
from PIL import Image

from .baidu import BaiduStreetView
from .google import GoogleStreetView

__all__ = ["Equirectangular"]


class Equirectangular:
    """
    用于将等距柱状全景图分割成普通透视图

    python tool to split equirectangular panorama into normal perspective view.

    修改自(Modified From)：https://github.com/fuenwang/Equirec2Perspec/tree/master

    MIT License

    Copyright (c) 2021 Fu-En Wang

    Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
    """

    def __init__(self, sv: Union[BaiduStreetView, GoogleStreetView]):
        self._img = self.pil2cv(sv.panorama)
        self._heading = sv.heading
        [self._height, self._width, _] = self._img.shape

    @staticmethod
    def pil2cv(img: Image.Image):
        """
        PIL.Image.Image -> OpenCV
        """
        cv2_img = np.array(img)
        cv2_img = cv2.cvtColor(cv2_img, cv2.COLOR_RGB2BGR)
        return cv2_img

    @staticmethod
    def cv2pil(img: np.ndarray):
        """
        OpenCV -> PIL.Image.Image
        """
        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        pil_img = Image.fromarray(img)
        return pil_img

    @staticmethod
    def _xyz2lonlat(xyz):
        atan2 = np.arctan2
        asin = np.arcsin

        norm = np.linalg.norm(xyz, axis=-1, keepdims=True)
        xyz_norm = xyz / norm
        x = xyz_norm[..., 0:1]
        y = xyz_norm[..., 1:2]
        z = xyz_norm[..., 2:]

        lon = atan2(x, z)
        lat = asin(y)
        lst = [lon, lat]

        out = np.concatenate(lst, axis=-1)
        return out

    @staticmethod
    def _lonlat2XY(lonlat, shape):
        X = (lonlat[..., 0:1] / (2 * np.pi) + 0.5) * (shape[1] - 1)
        Y = (lonlat[..., 1:] / (np.pi) + 0.5) * (shape[0] - 1)
        lst = [X, Y]
        out = np.concatenate(lst, axis=-1)

        return out

    def get_perspective(
        self, fov: float, heading: float, pitch: float, height: float, width: float
    ):
        """
        全景图透视投影
        Equirectangular to Perspective

        Args:
        - fov: 视野（单位：度）。Field of View (unit: degree)
        - heading: 水平方向朝向角（单位：度），范围为[0, 360]，正东方向为0度，逆时针旋转。Horizontal direction angle (unit: degree), range is [0, 360], 0 degrees is east, counterclockwise rotation.
        - pitch: 俯仰角（单位：度），范围为[-90, 90]，0度为水平方向，90度为垂直向上，-90度为垂直向下。Pitch angle (unit: degree), range is [-90, 90], 0 degrees is the horizontal direction, 90 degrees is vertical up, and -90 degrees is vertical down.
        - height: 输出图像高度。Output image height.
        - width: 输出图像宽度。Output image width.
        """
        heading = -(self._heading + heading + 180)

        f = 0.5 * width * 1 / np.tan(0.5 * fov / 180.0 * np.pi)
        cx = (width - 1) / 2.0
        cy = (height - 1) / 2.0
        K = np.array(
            [
                [f, 0, cx],
                [0, f, cy],
                [0, 0, 1],
            ],
            np.float32,
        )
        K_inv = np.linalg.inv(K)

        x = np.arange(width)
        y = np.arange(height)
        x, y = np.meshgrid(x, y)
        z = np.ones_like(x)
        xyz = np.concatenate([x[..., None], y[..., None], z[..., None]], axis=-1)
        xyz = xyz @ K_inv.T

        y_axis = np.array([0.0, 1.0, 0.0], np.float32)
        x_axis = np.array([1.0, 0.0, 0.0], np.float32)
        R1, _ = cv2.Rodrigues(y_axis * np.radians(heading))
        R2, _ = cv2.Rodrigues(np.dot(R1, x_axis) * np.radians(pitch))
        R = R2 @ R1
        xyz = xyz @ R.T
        lonlat = self._xyz2lonlat(xyz)
        XY = self._lonlat2XY(lonlat, shape=self._img.shape).astype(np.float32)
        persp = cv2.remap(
            self._img,
            XY[..., 0],
            XY[..., 1],
            cv2.INTER_CUBIC,
            borderMode=cv2.BORDER_WRAP,
        )

        return self.cv2pil(persp)



import geopandas as gpd
import numpy as np
from shapely.geometry import Polygon


def get_one_point(shp):
    """
    get one point from the shapefile
    """
    first_geometry = shp.geometry.iloc[0]

    # for different types of geometry, get the first point
    if first_geometry.geom_type == "Polygon":
        first_point = first_geometry.exterior.coords[0]
    elif first_geometry.geom_type == "MultiPolygon":
        first_polygon = list(first_geometry.geoms)[0]
        first_point = first_polygon.exterior.coords[0]
    else:
        raise ValueError("Geometry type not supported")

    pointx, pointy = first_point[0], first_point[1]

    return pointx, pointy


def deg2XY(lon_deg, lat_deg, zoom=15):
    """
    The satellite images are tiles and each tile has a unique x and y index.
    """
    # convert the latitude to radians
    lat_rad = np.radians(lat_deg)

    # the total number of the tiles in the x and y direction
    n = 2.0**zoom

    # the x index of the tile
    xtile = int((lon_deg + 180.0) / 360.0 * n)
    # the y index of the tile
    ytile = int(
        (1.0 - np.log(np.tan(lat_rad) + (1 / np.cos(lat_rad))) / np.pi) / 2.0 * n
    )

    return xtile, ytile


def XY2deg(x, y, zoom=15):
    # the total number of the tiles in the x and y direction
    n = 2.0**zoom

    # the longitude of the tile
    lon_deg = x / n * 360.0 - 180.0
    # the latitude of the tile
    lat_rad = np.arctan(np.sinh(np.pi * (1 - 2 * y / n)))
    # convert the latitude to degrees
    lat_deg = np.rad2deg(lat_rad)

    return lat_deg, lon_deg


def compute_tile_coordinates(min_x, max_x, min_y, max_y):
    """
    Batch version of YX2deg
    """
    x_arr = np.arange(min_x, max_x + 1)
    y_arr = np.arange(min_y, max_y + 1)
    lon_arr, lat_arr = XY2deg_batch(x_arr, y_arr)

    return lon_arr, lat_arr, x_arr, y_arr


def XY2deg_batch(x_arr, y_arr, zoom=15):
    """
    Batch version of num2deg
    """
    n = 2.0**zoom
    lon_deg_arr = x_arr / n * 360.0 - 180.0
    lat_rad_arr = np.arctan(np.sinh(np.pi * (1 - 2 * y_arr / n)))
    lat_deg_arr = np.rad2deg(lat_rad_arr)

    return lon_deg_arr, lat_deg_arr


def create_tile_polygons(lon_arr, lat_arr, x_arr, y_arr):
    """
    Create the polygons for each tile
    """
    # create the lon and lat meshgrid
    lon_mesh, lat_mesh = np.meshgrid(lon_arr, lat_arr, indexing="ij")
    x_mesh, y_mesh = np.meshgrid(x_arr, y_arr, indexing="ij")

    # create the polygons
    vertices = np.array(
        [
            lon_mesh[:-1, :-1],
            lat_mesh[:-1, :-1],
            lon_mesh[1:, :-1],
            lat_mesh[1:, :-1],
            lon_mesh[1:, 1:],
            lat_mesh[1:, 1:],
            lon_mesh[:-1, 1:],
            lat_mesh[:-1, 1:],
        ]
    )
    vertices = vertices.reshape(4, 2, -1)
    vertices = np.transpose(vertices, axes=(2, 0, 1))
    polygons = [Polygon(p) for p in vertices]

    # create the x and y coordinates
    vertices_x_y = np.array(
        [
            x_mesh[:-1, :-1],
            y_mesh[:-1, :-1],
            x_mesh[1:, :-1],
            y_mesh[1:, :-1],
            x_mesh[1:, 1:],
            y_mesh[1:, 1:],
            x_mesh[:-1, 1:],
            y_mesh[:-1, 1:],
        ]
    )
    vertices_x_y = vertices_x_y.reshape(4, 2, -1)
    vertices_x_y = np.transpose(vertices_x_y, axes=(2, 0, 1))
    y_x = [f"{int(p[0][1])}_{int(p[0][0])}" for p in vertices_x_y]

    # create the GeoDataFrame
    tile_gpd = gpd.GeoDataFrame({"Y_X": y_x}, geometry=polygons, crs="EPSG:4326")

    return tile_gpd


def geometry_to_listXY(geometry):
    """
    Convert the geometries to a list of x and y coordinates,
    which can be used as the index to download the satellite images (tiles).
    """
    # get the bbox of the geometries
    """
    may need to fix, when the box cross the 0 degree longitude
    """
    # get the bbox of the geometries
    min_x, min_y, max_x, max_y = geometry.bounds
    minx, maxy = deg2XY(min_x, min_y)
    maxx, miny = deg2XY(max_x, max_y)

    # compute the coordinates of the tiles
    lon_arr, lat_arr, x_arr, y_arr = compute_tile_coordinates(minx, maxx, miny, maxy)
    tile_gpd = create_tile_polygons(lon_arr, lat_arr, x_arr, y_arr)

    # get the tiles that intersect with the geometries
    intersection = gpd.sjoin(
        tile_gpd,
        gpd.GeoDataFrame(data={"tmp": [0]}, geometry=[geometry], crs="EPSG:4326").drop(
            columns=["tmp"]
        ),
        predicate="intersects",
        how="inner",
    )
    Y_X = list(intersection.Y_X)

    # if there is only one tile, then return all the tiles
    if len(Y_X) < 2:
        for x in range(minx, maxx + 1):
            for y in range(miny, maxy + 1):
                Y_X.append(f"{y}_{x}")

    return Y_X


def get_YX_area(area_shp):
    """
    Get the x and y coordinates of the tiles for the given area.
    """
    # unify the Coordinate Reference System (CRS) of the geometries
    area_shp = area_shp.to_crs(epsg=4326)

    # get the coordinates of the tiles
    geometries = area_shp["geometry"]
    area_shp["Y_X"] = geometries.map(geometry_to_listXY)

    # remove the duplicate tiles
    Y_X = sum(list(area_shp["Y_X"]), [])
    Y_X = sorted(list(set(Y_X)))

    return area_shp, Y_X



from pathlib import Path
import json
import re
from typing import List, Optional, Dict, Union
import itertools
import pickle
from io import BytesIO

import requests
from dataclasses import dataclass
from PIL import Image

__all__ = ["Panorama", "GoogleStreetView"]


@dataclass
class Panorama:
    pano_id: str
    lat: float
    lon: float
    heading: float
    pitch: Optional[float]
    roll: Optional[float]
    date: Optional[str]


class GoogleStreetView:

    def __init__(
        self,
        meta: Panorama,
        zoom: int = 4,
        headers: Optional[Dict[str, str]] = None,
        proxies: Optional[Dict[str, str]] = None,
    ):
        self.meta: Panorama = meta
        self._zoom = zoom
        self._headers = headers
        self._proxies = proxies
        self._raw_panorama, self.panorama = self._query_img()

    @property
    def heading(self) -> float:
        """
        获取拍摄时的朝向（单位：度）（统一到百度的heading计算方式上）
        Get the direction when the photo was taken (unit: degree) (unified to the heading calculation method of Baidu)
        """
        return self.meta.heading + 90

    @staticmethod
    def _extract_meta(text: str) -> List[Panorama]:
        """
        从panoids端点的有效响应中返回所有全景图的列表。
        Given a valid response from the panoids endpoint, return a list of all the
        panoids.
        """

        # The response is actually javascript code. It's a function with a single
        # input which is a huge deeply nested array of items.
        blob = re.findall(r"callbackfunc\( (.*) \)$", text)[0]
        data = json.loads(blob)

        if data == [[5, "generic", "Search returned no images."]]:
            return []

        subset = data[1][5][0]

        raw_panos = subset[3][0]

        if len(subset) < 9 or subset[8] is None:
            raw_dates = []
        else:
            raw_dates = subset[8]

        # For some reason, dates do not include a date for each panorama.
        # the n dates match the last n panos. Here we flip the arrays
        # so that the 0th pano aligns with the 0th date.
        raw_panos = raw_panos[::-1]
        raw_dates = raw_dates[::-1]

        dates = [f"{d[1][0]}-{d[1][1]:02d}" for d in raw_dates]

        return [
            Panorama(
                pano_id=pano[0][1],
                lat=pano[2][0][2],
                lon=pano[2][0][3],
                heading=pano[2][2][0],
                pitch=pano[2][2][1] if len(pano[2][2]) >= 2 else None,
                roll=pano[2][2][2] if len(pano[2][2]) >= 3 else None,
                date=dates[i] if i < len(dates) else None,
            )
            for i, pano in enumerate(raw_panos)
        ]

    @staticmethod
    def search(
        lng: float,
        lat: float,
        zoom: int = 4,
        headers: Optional[Dict[str, str]] = None,
        proxies: Optional[Dict[str, str]] = None,
        cache_dir: Union[str, Path, None] = None,
    ) -> "GoogleStreetView":
        """
        获取最接近GPS坐标的全景图（ids）。
        Gets the closest panoramas (ids) to the GPS coordinates.

        Args:
        - lng: 经度 (WGS84)。Longitude (WGS84).
        - lat: 纬度 (WGS84)。Latitude (WGS84).
        - zoom: 缩放级别，越大图像质量越高，最大为4（可选，默认为4）。Zoom level, the larger the image quality, the higher the maximum is 4 (optional, default is 4).
        - proxies: 代理（可选）。Proxies (optional).
        - cache_dir: 缓存目录，若为None则不缓存。Cache directory, if None, no cache.

        Returns:
        - sv: 街景图片。Street view image.
        """
        url = (
            "https://maps.googleapis.com/maps/api/js/"
            "GeoPhotoService.SingleImageSearch"
            f"?pb=!1m5!1sapiv3!5sUS!11m2!1m1!1b0!2m4!1m2!3d{lat}!4d{lng}!2d50!3m10"
            "!2m2!1sen!2sGB!9m1!1e2!11m4!1m3!1e2!2b1!3e2!4m10!1e1!1e2!1e3!1e4"
            "!1e8!1e6!5m1!1e2!6m1!1e2"
            "&callback=callbackfunc"
        )
        resp = requests.get(url, proxies=proxies)
        pans = GoogleStreetView._extract_meta(resp.text)
        if len(pans) == 0:
            raise ValueError("No panoramas found.")
        pan = pans[0]
        pid = pan.pano_id

        # 检查缓存
        if cache_dir is not None:
            if isinstance(cache_dir, str):
                cache_dir = Path(cache_dir)
            # 检查缓存目录是否存在，不存在则创建
            if not cache_dir.exists():
                cache_dir.mkdir(parents=True)
            cache_file = cache_dir / f"{pid}.pkl"
        else:
            cache_file = None
        if cache_file is not None and cache_file.exists():
            with open(cache_file, "rb") as f:
                sv = pickle.load(f)
        else:
            sv = GoogleStreetView(pan, zoom, headers=headers, proxies=proxies)
            if cache_file is not None:
                with open(cache_file, "wb") as f:
                    pickle.dump(sv, f)

        return sv

    def _query_img(self):
        """
        下载Google全景图
        Downloads a streetview panorama.
        """

        tile_width = 512
        tile_height = 512

        total_width, total_height = 2**self._zoom, 2 ** (self._zoom - 1)
        raw = Image.new("RGB", (total_width * tile_width, total_height * tile_height))

        with requests.Session() as s:
            for x, y in itertools.product(range(total_width), range(total_height)):
                url = f"https://streetviewpixels-pa.googleapis.com/v1/tile?cb_client=maps_sv.tactile&panoid={self.meta.pano_id}&x={x}&y={y}&zoom={self._zoom}"  # 可以增加参数nbt=1（no black tile），但是会导致部分图片返回400错误
                with s.get(
                    url,
                    headers=self._headers,
                    proxies=self._proxies,
                    stream=True,
                ) as resp:
                    if resp.status_code != 200:
                        print(resp.json())
                        continue
                    b = resp.content
                    img = Image.open(BytesIO(b))
                    raw.paste(im=img, box=(x * tile_width, y * tile_height))
        # 移除黑边
        return raw, raw.crop(raw.getbbox())



from pathlib import Path
import pickle
from io import BytesIO
from typing import Dict, Optional, Union

import requests
from PIL import Image

__all__ = ["BaiduStreetView"]


class BaiduStreetView:
    def __init__(
        self,
        sid: str,
        zoom: int = 4,
        headers: Optional[Dict[str, str]] = None,
        proxies: Optional[Dict[str, str]] = None,
    ):
        self.sid = sid
        self._zoom = zoom
        self._headers = headers
        self._proxies = proxies
        self.meta = self._query_camera()
        """
        设备信息，关键字段如下（用于调整取出照片时的方位）：
        Camera Information, the useful fields are as follows (used to adjust the orientation of the taken photo):
        - "Heading": 拍摄时的朝向（单位：度），全景相片的0度方向总是为拍摄时朝向的左侧，90度为正前方，180度为拍摄时朝向的右侧，270度为正后方。The direction of the photo taken (unit: degree), the 0-degree direction of the panorama photo is always to the left of the direction when the photo was taken, 90 degrees is the front, 180 degrees is to the right of the direction when the photo was taken, and 270 degrees is the back.
        - "Pitch": 拍摄时的俯仰角（单位：度），范围为[-90, 90]，0度为水平方向，90度为垂直向上，-90度为垂直向下。The pitch angle when the photo was taken (unit: degree), the range is [-90, 90], 0 degrees is the horizontal direction, 90 degrees is vertical up, and -90 degrees is vertical down.
        - "X"/"Y"/"Z": 拍摄时的坐标（单位：米），BD09MC坐标系。The coordinates when the photo was taken (unit: meter), BD09MC coordinate system.
        - "Rname": 拍摄时的道路名称。The road name when the photo was taken.
        """
        self.panorama = self._query_img()
        """
        全景图，采用Equirectangular转换为普通透视图后可用于显示。
        Panorama, which can be used for display after being converted to a normal perspective view using Equirectangular.
        """

    @property
    def heading(self) -> float:
        """
        获取拍摄时的朝向（单位：度）
        Get the direction when the photo was taken (unit: degree)
        """
        return self.meta["Heading"]

    @staticmethod
    def search(
        x,
        y,
        headers: Optional[Dict[str, str]] = None,
        proxies: Optional[Dict[str, str]] = None,
        cache_dir: Union[str, Path, None] = None,
    ) -> "BaiduStreetView":
        """
        街景图片信息查询
        Street view image information query

        Args:
        - x: x坐标（BD09MC）。x coordinate (BD09MC)
        - y: y坐标（BD09MC）。y coordinate (BD09MC)
        - cache_dir: 缓存目录，若为None则不缓存。Cache directory, if None, no cache.

        Returns:
        - sv: 街景图片。Street view image.
        """
        url = f"https://mapsv0.bdimg.com/?qt=qsdata&x={x}&y={y}"
        resp = requests.get(url)
        result = resp.json()
        # 正确响应格式：{'content': {'RoadId': 'c68cdd-64ba-7b9a-549d-82eac9', 'RoadName': '锡拉胡同', 'id': '09002200121707120825100452M', 'x': 1295957900, 'y': 482703000}, 'result': {'action': 0, 'error': 0}}
        # 失败响应格式：{'result': {'action': 0, 'error': 400}}
        if result["result"]["error"] != 0:
            raise ValueError(f"invalid qsdata response: {result}")
        sid = result["content"]["id"]

        # 检查缓存
        if cache_dir is not None:
            if isinstance(cache_dir, str):
                cache_dir = Path(cache_dir)
            # 检查缓存目录是否存在，不存在则创建
            if not cache_dir.exists():
                cache_dir.mkdir(parents=True)
            cache_file = cache_dir / f"{sid}.pkl"
        else:
            cache_file = None
        if cache_file is not None and cache_file.exists():
            with open(cache_file, "rb") as f:
                sv = pickle.load(f)
        else:
            sv = BaiduStreetView(sid, headers=headers, proxies=proxies)
            if cache_file is not None:
                with open(cache_file, "wb") as f:
                    pickle.dump(sv, f)
        return sv

    def _query_camera(self):
        url = f"https://mapsv0.bdimg.com/?qt=sdata&sid={self.sid}"
        resp = requests.get(url)
        result = resp.json()
        if result["result"]["error"] != 0 or len(result["content"]) != 1:
            raise ValueError(f"invalid sdata response: {result}")
        return result["content"][0]

    def _query_img(self):
        # 1. 下载全景图
        if self._zoom == 1:
            xrange, yrange = 1, 1
        elif self._zoom == 2:
            xrange, yrange = 1, 2
        elif self._zoom == 3:
            xrange, yrange = 2, 4
        elif self._zoom == 4:
            xrange, yrange = 4, 8
        else:
            raise ValueError(f"invalid zoom: {self._zoom}")
        imgs = []
        with requests.Session() as s:
            for x in range(xrange):
                for y in range(yrange):
                    url = f"https://mapsv0.bdimg.com/?qt=pdata&sid={self.sid}&pos={x}_{y}&z={self._zoom}&from=PC"
                    with s.get(url) as resp:
                        b = resp.content
                        img = Image.open(BytesIO(b))
                        imgs.append(img)

        # 2. 拼接全景图
        w, h = imgs[0].size
        width = w * yrange
        height = h * xrange
        panorama = Image.new("RGB", (width, height))
        for i, img in enumerate(imgs):
            x = i // yrange
            y = i % yrange
            panorama.paste(
                img,
                (
                    y * h,
                    x * w,
                ),
            )
        return panorama



"""
街景图片操作库
Street View Image Library
"""

from typing import List, Tuple
import requests

from .equirectangular import Equirectangular
from .baidu import BaiduStreetView
from .google import GoogleStreetView

__all__ = ["BaiduStreetView", "GoogleStreetView", "Equirectangular", "wgs842bd09mc"]


def wgs842bd09mc(
    coords: List[Tuple[float, float]], ak: str
) -> List[Tuple[float, float]]:
    """
    坐标转换：WGS84 -> BD09MC
    Coordinate transformation: WGS84 -> BD09MC

    Args:
    - coords: 需转换的源坐标，多组坐标以“;”分隔（经度,纬度）。The source coordinates to be converted, separated by ";" (longitude, latitude).
    - ak: 百度地图开发者密钥。Baidu Map Developer Key.

    Returns:
    - points: 转换后的坐标。The converted coordinates.
    """
    if len(coords) > 100:
        raise ValueError("coords too many")
    coordstr = ";".join([f"{x[0]},{x[1]}" for x in coords])
    url = f"http://api.map.baidu.com/geoconv/v1/?coords={coordstr}&from=1&to=6&ak={ak}"
    resp = requests.get(url)
    result = resp.json()
    xys = [(one["x"], one["y"]) for one in result["result"]]
    return xys
